#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "aiot_device_api.h"
#include "aiot_dm_api.h"
#include "demo_config.h"

static void demo_status_callback(void *device, const aiot_device_status_t *status, void *userdata)
{
    switch (status->type) {
    /* SDK因为用户调用了aiot_device_connect()接口, 与mqtt服务器建立连接已成功 */
    case AIOT_DEVICE_STATUS_CONNECT: {
        printf("AIOT_DEVICE_STATUS_CONNECT\n");
        /* TODO: 处理SDK建连成功, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;

    /* SDK因为网络状况被动断连后, 自动发起重连已成功 */
    case AIOT_DEVICE_STATUS_RECONNECT: {
        printf("AIOT_DEVICE_STATUS_CONNECT\n");
        /* TODO: 处理SDK重连成功, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;

    /* SDK因为网络的状况而被动断开了连接, network是底层读写失败, heartbeat是没有按预期得到服务端心跳应答 */
    case AIOT_DEVICE_STATUS_DISCONNECT: {
        printf("AIOT_DEVICE_STATUS_CONNECT: %d\n", status->error_code);
        /* TODO: 处理SDK被动断连, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;
    }
}

/* 设备的消息回调示例，默认只做打印处理 */
static void demo_msg_callback(void *device, const aiot_msg_t *message, void *userdata)
{
    /* 设备接收到所有消息都会执行该回调，如果只使用物模型功能可以不实现 */
}

/* 云端->设备：通用的云端回复的消息处理示例, 示例只做打印处理 */
static void demo_dm_msg_generic_reply_callback(void *device, const aiot_dm_msg_t *dm_msg, void *userdata)
{
    printf("demo_dm_recv_generic_reply code = %d, data = %.*s, message = %.*s\r\n",
        dm_msg->data.generic_reply.code,
        dm_msg->data.generic_reply.data_len,
        dm_msg->data.generic_reply.data,
        dm_msg->data.generic_reply.message_len,
        dm_msg->data.generic_reply.message);
}

/* 云端->设备：属性设置消息处理示例, 示例只做打印处理  */
static void demo_dm_msg_property_set_callback(void *device, const aiot_dm_msg_t *dm_msg, void *userdata)
{
    printf("demo_dm_msg_property_set_callback params = %.*s\r\n",
        dm_msg->data.property_set.params_len,
        dm_msg->data.property_set.params);

    /* TODO 演示同步回复 */
    char *reply_data = "{}";
    aiot_device_dm_propertyset_reply(device, dm_msg->context, 200, reply_data);

    /* TODO 演示异步回复[不在回调函数中回复], 需要备份消息的上下文用于回复 */
    /*
    void *context = aiot_dm_msg_context_clone(dm_msg->context);
    aiot_device_dm_propertyset_reply(device, dm_msg->context, 200, reply_data);
    aiot_dm_msg_context_free(context);
    */
}

/* 云端->设备：服务调用消息处理示例, 示例只做打印处理  */
static void demo_dm_msg_service_invoke_callback(void *device, const aiot_dm_msg_t *dm_msg, void *userdata)
{
    printf("demo_dm_msg_service_invoke_callback service_id %s, params = %.*s\r\n",
        dm_msg->data.service_invoke.service_id,
        dm_msg->data.service_invoke.params_len,
        dm_msg->data.service_invoke.params);

    /* TODO 演示同步回复 */
    char *reply_data = "{}";
    aiot_device_dm_service_reply(device, dm_msg->context, dm_msg->data.service_invoke.service_id, 200, reply_data);

    /* TODO 演示异步回复[不在回调函数中回复], 需要备份消息的上下文用于回复 */
    /*
    void *context = aiot_dm_msg_context_clone(dm_msg->context);
    aiot_device_dm_propertyset_reply(device, dm_msg->context, 200, reply_data);
    aiot_dm_msg_context_free(context);
    */
}

/* 物模型消息的回调函数 */
static void demo_dm_msg_callback(void *device, const aiot_dm_msg_t *dm_msg, void *userdata)
{
    if(dm_msg == NULL) {
        return;
    }
    printf("[thing-message] recv msg type %d\r\n", dm_msg->type);
    switch(dm_msg->type){
        case AIOT_DMRECV_POST_REPLY:
        case AIOT_DMRECV_BATCH_POST_REPLY:
        case AIOT_DMRECV_GET_DESIRED_REPLY:
        case AIOT_DMRECV_DELETE_DESIRED_REPLY:
            demo_dm_msg_generic_reply_callback(device, dm_msg, userdata);
        break;
        case AIOT_DMRECV_PROPERTY_SET:
            demo_dm_msg_property_set_callback(device, dm_msg, userdata);
        break;
        case AIOT_DMRECV_SERVICE_INVOKE:
            demo_dm_msg_service_invoke_callback(device, dm_msg, userdata);
        break;
        default:
            break;
    }
}

static void* demo_device_init(char *product_key, char *device_name, char *device_secret, char *host, uint16_t port)
{
    int32_t res = STATE_SUCCESS;
    /* 创建设备 */
    void *device = aiot_device_create(product_key, device_name);
    if (device == NULL) {
        printf("device create failed\n");
        return NULL;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(device, device_secret);

    /* 连接配置参数初始化 */
    aiot_linkconfig_t* config = aiot_linkconfig_init(protocol);
    /* 设置服务器的host、port */
    aiot_linkconfig_host(config, host, port);

    /* 设置设备连接参数 */
    aiot_device_set_linkconfig(device, config);
    
    /* 设备建连 */
    res = aiot_device_connect(device);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_linkconfig_deinit(&config);
        aiot_device_delete(&device);
        printf("aiot_device_connect failed: -0x%04X\n\r\n", -res);
        return NULL;
    }

    /* 建连成功返回设备对象 */
    aiot_linkconfig_deinit(&config);
    return device;
}

static void demo_device_deinit(void *device)
{
    int32_t res = STATE_SUCCESS;
    res = aiot_device_disconnect(device);
    if (res < STATE_SUCCESS) {
        printf("aiot_device_disconnect failed: -0x%04X\n", -res);
    }
    aiot_device_delete(&device);
}

/*
    TODO: 运行前先编辑demo_config.h配置demo使用的设备信息及建连信息
*/
int main(int argc, char *argv[])
{
    /* 配置SDK的底层依赖 */
    aiot_sysdep_init(aiot_sysdep_get_portfile());

    /* 初始化设备，并将设备连接上云 */
    void *device_client = demo_device_init(product_key, device_name, device_secret, host, port);
    if(device_client == NULL) {
        printf("demo device connect failed\r\n");
        return -1;
    }

    /* 设置设备消息(对应所有消息)回调及状态变化回调 */
    aiot_device_set_event_callback(device_client, demo_msg_callback, demo_status_callback, NULL, NULL);
    /* 设置物模型消息回调 */
    aiot_device_dm_set_msg_callback(device_client, demo_dm_msg_callback, NULL);

    /* 获取期望属性 */
    /*
    char *get_params = "[\"LightSwitch\"]";
    aiot_device_dm_get_desired(device_client, get_params);
    */

    /* 删除期望属性 */
    /*
    char *delete_desired_params = "{\"LightSwitch\":{}}";
    aiot_device_dm_delete_desired(device_client, delete_desired_params);
    */

    while(1) {
        /* 属性上报示例 */
        char *property_params = "{\"LightSwitch\": 0}";
        aiot_device_dm_property_post(device_client, property_params, 1);
        sleep(1);

        /* 事件上报示例 */
        char *event_id = "Error";
        char *event_params = "{\"ErrorCode\": -1}";
        aiot_device_dm_event_post(device_client, event_id, event_params, 1);
        sleep(1);

        /* 批量属性上报，可用于上报属性的时序数据 */
        /*
        char *batchparams = "{\"properties\":{\"Power\": [ {\"value\":\"on\",\"time\":1612684518}],\"WF\": [{\"value\": 3,\"time\":1612684518}]}}";
        aiot_device_dm_batch_post(device_client, batchparams, 1);
        sleep(1);
        */
    }

    /* 断开设备连接，并回收设备资源 */
    demo_device_deinit(device_client);
    return 0;
}