#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "aiot_device_api.h"
#include "aiot_dm_api.h"
#include "demo_config.h"

static void demo_status_callback(void *device, const aiot_device_status_t *status, void *userdata)
{
    switch (status->type) {
    /* SDK因为用户调用了aiot_device_connect()接口, 与mqtt服务器建立连接已成功 */
    case AIOT_DEVICE_STATUS_CONNECT: {
        printf("AIOT_DEVICE_STATUS_CONNECT\n");
        /* TODO: 处理SDK建连成功, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;

    /* SDK因为网络状况被动断连后, 自动发起重连已成功 */
    case AIOT_DEVICE_STATUS_RECONNECT: {
        printf("AIOT_DEVICE_STATUS_CONNECT\n");
        /* TODO: 处理SDK重连成功, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;

    /* SDK因为网络的状况而被动断开了连接, network是底层读写失败, heartbeat是没有按预期得到服务端心跳应答 */
    case AIOT_DEVICE_STATUS_DISCONNECT: {
        printf("AIOT_DEVICE_STATUS_CONNECT: %d\n", status->error_code);
        /* TODO: 处理SDK被动断连, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;
    }
}

/* 设备的消息回调示例，默认只做打印处理 */
static void demo_msg_callback(void *device, const aiot_msg_t *message, void *userdata)
{
    /* 设备接收到所有消息都会执行该回调，如果只使用物模型功能可以不实现 */
}

/* 云端->设备：透传的消息处理示例, 示例只做打印处理  */
static void demo_dm_msg_down_raw_callback(void *device, const aiot_dm_msg_t *dm_msg, void *userdata)
{
    printf("demo_dm_msg_down_raw_callback raw data len = %d\r\n", dm_msg->data.raw_data.data_len);
}

/* 物模型消息的回调函数 */
static void demo_dm_msg_callback(void *device, const aiot_dm_msg_t *dm_msg, void *userdata)
{
    if(dm_msg == NULL) {
        return;
    }
    printf("[thing-message] recv msg type %d\r\n", dm_msg->type);
    switch(dm_msg->type){
        case AIOT_DMRECV_RAW_REPLY:
        case AIOT_DMRECV_RAW_DOWN:
            demo_dm_msg_down_raw_callback(device, dm_msg, userdata);
        break;
        default:
            break;
    }
}

static void* demo_device_init(char *product_key, char *device_name, char *device_secret, char *host, uint16_t port)
{
    int32_t res = STATE_SUCCESS;
    /* 创建设备 */
    void *device = aiot_device_create(product_key, device_name);
    if (device == NULL) {
        printf("device create failed\n");
        return NULL;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(device, device_secret);

    /* 连接配置参数初始化 */
    aiot_linkconfig_t* config = aiot_linkconfig_init(protocol);
    /* 设置服务器的host、port */
    aiot_linkconfig_host(config, host, port);

    /* 设置设备连接参数 */
    aiot_device_set_linkconfig(device, config);
    
    /* 设备建连 */
    res = aiot_device_connect(device);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_linkconfig_deinit(&config);
        aiot_device_delete(&device);
        printf("aiot_device_connect failed: -0x%04X\n\r\n", -res);
        return NULL;
    }

    /* 建连成功返回设备对象 */
    aiot_linkconfig_deinit(&config);
    return device;
}

static void demo_device_deinit(void *device)
{
    int32_t res = STATE_SUCCESS;
    res = aiot_device_disconnect(device);
    if (res < STATE_SUCCESS) {
        printf("aiot_device_disconnect failed: -0x%04X\n", -res);
    }
    aiot_device_delete(&device);
}

/*
    TODO: 运行前先编辑demo_config.h配置demo使用的设备信息及建连信息
*/
int main(int argc, char *argv[])
{
    /* 配置SDK的底层依赖 */
    aiot_sysdep_init(aiot_sysdep_get_portfile());

    /* 初始化设备，并将设备连接上云 */
    void *device_client = demo_device_init(product_key, device_name, device_secret, host, port);
    if(device_client == NULL) {
        printf("demo device connect failed\r\n");
        return -1;
    }

    /* 设置设备消息(对应所有消息)回调及状态变化回调 */
    aiot_device_set_event_callback(device_client, demo_msg_callback, demo_status_callback, NULL, NULL);
    /* 设置物模型消息回调 */
    aiot_device_dm_set_msg_callback(device_client, demo_dm_msg_callback, NULL);

    while(1) {
        /* 上报二进制数据 */
        uint8_t raw_data[] = { 0x01, 0x02 };
        aiot_device_dm_raw_post(device_client, raw_data, sizeof(raw_data));
        sleep(5);
    }

    /* 断开设备连接，并回收设备资源 */
    demo_device_deinit(device_client);
    return 0;
}