#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "aiot_device_api.h"
#include "demo_config.h"

void demo_status_callback(void *device, const aiot_device_status_t *status, void *userdata)
{
    /* TODO: 设备状态变化的回调函数，用户可根据业务需求执行对应业务，回调不可长时间阻塞, 回调中不可调用设备删除操作 */
    switch (status->type) {
    /* SDK因为用户调用了aiot_device_connect()接口, 与服务器建立连接已成功 */
    case AIOT_DEVICE_STATUS_CONNECT: {
        printf("AIOT_DEVICE_STATUS_CONNECT\n");
    }
    break;

    /* SDK因为网络状况被动断连后, 自动发起重连已成功 */
    case AIOT_DEVICE_STATUS_RECONNECT: {
        printf("AIOT_DEVICE_STATUS_CONNECT\n");
    }
    break;

    /* SDK因为网络的状况而被动断开了连接, network是底层读写失败, heartbeat是没有按预期得到服务端心跳应答 */
    case AIOT_DEVICE_STATUS_DISCONNECT: {
        printf("AIOT_DEVICE_STATUS_CONNECT: %d\n", status->error_code);
    }
    break;
    }
}


void demo_msg_callback(void *device, const aiot_msg_t *message, void *userdata)
{
    /* 设备接收到消息消息，默认只做打印处理 */
    printf("[message] <<, topic: %s\n", message->topic);
    printf("[message] <<, payload: %.*s\n", message->payload_lenth, message->payload);
}

void demo_result_callback(void *device, const aiot_device_msg_result_t *result, void *userdata)
{
    switch(result->type) {
        /* 发送qos消息，收到后执行的回调 */
        case AIOT_DEVICE_SEND_MESSAGE_RESULT: {
            printf("send message result, message id: %d, code -0x%04X\n", result->message_id, -result->code);
        }
        break;  
        /* 注册订阅topic后得到回复执行的回调 */
        case AIOT_DEVICE_REGISTER_TOPIC_RESULT:{
            printf("register, message id: %d, code code -0x%04X\n", result->message_id, -result->code);
        }
        break;  
        default:
            break;
    }
}

/* 自定义topic下行消息处理示例, 示例只做打印处理 */
void demo_user_topic_callback(void *device, const aiot_msg_t *message, void *userdata)
{
    printf("[thing_message] <<, topic: %s\n", message->topic);
    printf("[thing_message] <<, payload: %.*s\n", message->payload_lenth, message->payload);
}

int main(int argc, char *argv[])
{
    int32_t     res = STATE_SUCCESS;
    /* 配置SDK的底层依赖 */
    aiot_sysdep_init(aiot_sysdep_get_portfile());

    /* 创建设备 */
    void *device_client = aiot_device_create(product_key, device_name);
    if (device_client == NULL) {
        printf("device_client failed\n");
        return -1;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(device_client, device_secret);
    /* 设置设备消息回调及状态变化回调 */
    aiot_device_set_event_callback(device_client, demo_msg_callback, demo_status_callback, demo_result_callback, NULL);

    /* 连接配置参数初始化 */
    aiot_linkconfig_t* config = aiot_linkconfig_init(protocol);
    /* 设置服务器的host、port */
    aiot_linkconfig_host(config, host, port);

    /* 设置设备连接参数 */
    aiot_device_set_linkconfig(device_client, config);
    /* 设备建连 */
    res = aiot_device_connect(device_client);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_linkconfig_deinit(&config);
        aiot_device_delete(&device_client);
        printf("aiot_device_connect failed: -0x%04X\n\r\n", -res);
        return -1;
    }

    /* 订阅消息，设置该消息回调函数，也支持只设备回调 */
    char sub_topic[128] = { 0 };
    snprintf(sub_topic, sizeof(sub_topic), "/%s/%s/user/get", product_key, device_name);
    int32_t msg_id = aiot_device_register_topic_filter(device_client, sub_topic, demo_user_topic_callback, 1, NULL);
    if(msg_id >= 0){
        printf("aiot_device_register_topic_filter topic %s msg_id %d\r\n", sub_topic, msg_id);
    }
    
    while (1) {
        /* 创建消息 */
        char pub_topic[128] = { 0 };
        snprintf(pub_topic, sizeof(pub_topic), "/%s/%s/user/update", product_key, device_name);
        char *pub_payload = "{\"id\":\"1\",\"version\":\"1.0\",\"params\":{\"LightSwitch\":0}}";
        aiot_msg_t *pub_message = aiot_msg_create_raw(pub_topic, (uint8_t *)pub_payload, strlen(pub_payload));

        /* 发送消息 */
        aiot_device_send_message(device_client, pub_message);
        
        /* 删除消息 */
        aiot_msg_delete(pub_message);
        sleep(1);
    }

    res = aiot_device_disconnect(device_client);
    if (res < STATE_SUCCESS) {
        printf("aiot_device_disconnect failed: -0x%04X\n", -res);
        return -1;
    }
    /* 销毁设备实例, 一般不会运行到这里 */
    aiot_device_delete(&device_client);
    aiot_linkconfig_deinit(&config);
    return 0;
}
