#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "aiot_device_api.h"
#include "aiot_gateway_api.h"
#include "aiot_message_api.h"
#include "aiot_dm_api.h"
#include "demo_config.h"

/* 子设备的认证信息 */
aiot_subdev_meta_info_t sub_device_meta[] = SUBDEV_LIST;

/* 异步建连，返回结果标记*/
typedef struct {
    /* 子设备添加topo关系结果返回 */
    int add_topo_resp;
    /* 子设备批量建连结果返回 */
    int connect_resp;
    /* 子设备批量断连结果返回 */
    int disconnect_resp;
    /* 子设备删除topo关系结果返回*/
    int delete_topo_resp;
} demo_gateway_resp_flag_t;

/* 网关模块的消息回调 */
void demo_gateway_msg_callback(void *device, const aiot_gateway_msg_t *msg, void *userdata)
{
    demo_gateway_resp_flag_t *resp = (demo_gateway_resp_flag_t *)userdata;
    if(userdata == NULL) {
        return;
    }

    switch(msg->type){
        case AIOT_GWRECV_CONNECT_REPLY:{
            /* 子设备批量建连返回 */
            resp->connect_resp = msg->data.reply.code;
        }
        break;
        case AIOT_GWRECV_DISCONNECT_REPLY:{
            /* 子设备批量断连结果返回 */
            resp->disconnect_resp = msg->data.reply.code;
        }
        break;
        case AIOT_GWRECV_ADD_TOPO_REPLY:{
            /* 子设备添加topo关系结果返回 */
            resp->add_topo_resp = msg->data.reply.code;
        }
        break;
        case AIOT_GWRECV_DELETE_TOPO_REPLY:{
            /* 子设备删除topo关系结果返回 */
            resp->delete_topo_resp = msg->data.reply.code;
        }
        break;
        case AIOT_GWRECV_TOPO_CHANGE_NOTICE:{
            printf("type[%d] subdev_num %d status %d\r\n", msg->type, msg->subdev_num, msg->data.topo_change.status);
        }
        break;
    }
}

/* 设备的状态回调示例 */
void demo_status_callback(void *device, const aiot_device_status_t *status, void *userdata)
{
    char *product_key = aiot_device_product_key(device);
    char *device_name = aiot_device_name(device);
    if(device == NULL) {
        return;
    }

    switch (status->type) {
    /* SDK因为用户调用了aiot_device_connect()接口, 与mqtt服务器建立连接已成功 */
    case AIOT_DEVICE_STATUS_CONNECT: {
        printf("[%s][%s] AIOT_DEVICE_STATUS_CONNECT\n", product_key, device_name);
        /* TODO: 处理SDK建连成功, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;

    /* SDK因为网络状况被动断连后, 自动发起重连已成功 */
    case AIOT_DEVICE_STATUS_RECONNECT: {
        printf("[%s][%s] AIOT_DEVICE_STATUS_RECONNECT\n", product_key, device_name);
        /* TODO: 处理SDK重连成功, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;

    /* SDK因为网络的状况而被动断开了连接, network是底层读写失败, heartbeat是没有按预期得到服务端心跳应答 */
    case AIOT_DEVICE_STATUS_DISCONNECT: {
        printf("[%s][%s] AIOT_DEVICE_STATUS_DISCONNECT: %d\n", product_key, device_name, status->error_code);
        /* TODO: 处理SDK被动断连, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;
    }
}

/* 设备的消息回调示例，默认只做打印处理 */
void demo_msg_callback(void *device, const aiot_msg_t *message, void *userdata)
{
    char *product_key = aiot_device_product_key(device);
    char *device_name = aiot_device_name(device);
    if(device == NULL) {
        return;
    }

    printf("[%s][%s] <<, topic: %s\n", product_key, device_name, message->topic);
    printf("[%s][%s] <<, payload: %.*s\n", product_key, device_name, message->payload_lenth, message->payload);
}

/* 主动发送的消息，或者订阅/取消订阅消息有结果后会回调该消息 */
void demo_result_callback(void *device, const aiot_device_msg_result_t *result, void *userdata)
{
    char *product_key = aiot_device_product_key(device);
    char *device_name = aiot_device_name(device);
    if(device == NULL) {
        return;
    }

    switch(result->type) {
        case AIOT_DEVICE_SEND_MESSAGE_RESULT: {
            printf("[%s][%s] send message result, message id: %d, code -0x%04X\n", product_key, device_name, result->message_id, -result->code);
        }
        break;  
        case AIOT_DEVICE_REGISTER_TOPIC_RESULT:{
            printf("[%s][%s] register, message id: %d, code code -0x%04X\n", product_key, device_name, result->message_id, -result->code);
        }
        break;  
        default:
            break;
    }
}

static void* demo_gateway_device_init(char *product_key, char *device_name, char *device_secret, char *host, uint16_t port)
{
    int32_t res = STATE_SUCCESS;
    /* 创建设备 */
    void *device = aiot_device_create(product_key, device_name);
    if (device == NULL) {
        printf("device create failed\n");
        return NULL;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(device, device_secret);

    /* 连接配置参数初始化 */
    aiot_linkconfig_t* config = aiot_linkconfig_init(protocol);
    /* 设置服务器的host、port */
    aiot_linkconfig_host(config, host, port);

    /* 设置设备连接参数 */
    aiot_device_set_linkconfig(device, config);
    
    /* 设备建连 */
    res = aiot_device_connect(device);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_linkconfig_deinit(&config);
        aiot_device_delete(&device);
        printf("aiot_device_connect failed: -0x%04X\n\r\n", -res);
        return NULL;
    }

    /* 建连成功返回设备对象 */
    aiot_linkconfig_deinit(&config);
    return device;
}

static void demo_gateway_device_deinit(void *device)
{
    int32_t res = STATE_SUCCESS;
    res = aiot_device_disconnect(device);
    if (res < STATE_SUCCESS) {
        printf("aiot_device_disconnect failed: -0x%04X\n", -res);
    }
    aiot_device_delete(&device);
}

void *demo_sub_device_init(aiot_subdev_meta_info_t *device)
{
    /* 创建设备 */
    void *sub_device = aiot_device_create(device->product_key, device->device_name);
    if (sub_device == NULL) {
        printf("subdev_device failed\n");
        return NULL;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(sub_device, device->device_secret);
    /* 设置设备消息回调及状态变化回调
     * demo中子设备与网关设备使用的相同回调函数，最后的参数userdata可设置不同参数以做区分
     * */
    aiot_device_set_event_callback(sub_device, demo_msg_callback, demo_status_callback, demo_result_callback, NULL);
    return sub_device;
}

/* 删除子设备对象, 回收内存 */
void demo_sub_device_deinit(void **subdev_table, int32_t num)
{
    int i = 0;
    for(i = 0; i < num; i++) {
        aiot_device_delete(&subdev_table[i]);
    }

    free(subdev_table);
}

/* 等待超时 */
int demo_wait_expect_resp(int *value, int expect_value, int timeout_ms)
{
    aiot_sysdep_portfile_t *sysdep = aiot_sysdep_get_portfile();
    int origin_value = *value;
    uint64_t start_time = sysdep->core_sysdep_time();

    /* 等待返回值变化或者超时 */
    while(origin_value == *value && sysdep->core_sysdep_time() - start_time < timeout_ms) {
        sleep(1);
    }

    if( *value == expect_value ) {
        return 0;
    } else {
        return -1;
    }
}

int main(int argc, char *argv[])
{
    int32_t     res = STATE_SUCCESS;
    void **subdev_table = NULL;
    int32_t subdev_num = 0;
    /* 示例为10秒超时 */
    int32_t timeout_ms = 10 * 1000;
    int i = 0;
    demo_gateway_resp_flag_t resp;
    /* 配置SDK的底层依赖 */
    aiot_sysdep_init(aiot_sysdep_get_portfile());

    /* 网关设备完成初始化，启动建连，如果失败返回NULL */
    void *gateway_device = demo_gateway_device_init(gateway_product_key, gateway_device_name, gateway_device_secret, host, port);
    if(gateway_device == NULL) {
        return -1;
    }
    /* 设置网关消息回调 */
    memset(&resp, 0, sizeof(resp));
    aiot_gateway_set_msg_callback(gateway_device, demo_gateway_msg_callback, &resp);

    subdev_num = sizeof(sub_device_meta) / sizeof(aiot_subdev_meta_info_t);
    /* 子设备添加topo,异步调用 */
    res = aiot_gateway_batch_add_topo_async(gateway_device, sub_device_meta, subdev_num);
    if(res >= STATE_SUCCESS && 0 == demo_wait_expect_resp(&resp.add_topo_resp, 200, timeout_ms)) {
        printf("aiot_gateway_batch_add_topo_async success %d\r\n", res);
    } else {
        printf("aiot_gateway_batch_add_topo_async error %d\r\n", res);
        demo_gateway_device_deinit(gateway_device);
        return -1;
    }

    subdev_table = malloc(subdev_num * sizeof(void *));
    /* 子设备对象创建及初始化 */
    for(i = 0; i < subdev_num; i++) {
        subdev_table[i] = demo_sub_device_init(&sub_device_meta[i]);
    }

    /* 子设备建连，异步接口 */
    res = aiot_gateway_batch_connect_subdev_async(gateway_device, subdev_table, subdev_num);
    if(res >= STATE_SUCCESS && 0 == demo_wait_expect_resp(&resp.connect_resp, 200, timeout_ms)) {
        printf("aiot_gateway_batch_connect_subdev_async success %d\r\n", res);
    } else {
        printf("aiot_gateway_batch_connect_subdev_async error %d\r\n", res);
        demo_gateway_device_deinit(gateway_device);
        demo_sub_device_deinit(subdev_table, subdev_num);
        return -1;
    }

    while(1) {
        char *property_params = "{\"LightSwitch\": 0}";
        for(i = 0; i < sizeof(sub_device_meta) / sizeof(aiot_subdev_meta_info_t); i++) {
            /* 子设备发送物模型消息 */
            aiot_device_dm_property_post(subdev_table[i], property_params, 1);
        }
        /* 网关发送物模型消息 */
        aiot_device_dm_property_post(gateway_device, property_params, 1);
        sleep(1);
    }

    /* 子设备断开连接 */
    res = aiot_gateway_batch_disconnect_subdev_async(gateway_device, subdev_table, subdev_num);
    if(res >= STATE_SUCCESS && 0 == demo_wait_expect_resp(&resp.disconnect_resp, 200, timeout_ms)) {
        printf("aiot_gateway_batch_disconnect_subdev_async success %d\r\n", res);
    } else {
        printf("aiot_gateway_batch_disconnect_subdev_async error %d\r\n", res);
        demo_gateway_device_deinit(gateway_device);
        demo_sub_device_deinit(subdev_table, subdev_num);
        return -1;
    }

    /* 子设备删除topo */
    res = aiot_gateway_batch_delete_topo_async(gateway_device, sub_device_meta, subdev_num);
    if(res >= STATE_SUCCESS && 0 == demo_wait_expect_resp(&resp.delete_topo_resp, 200, timeout_ms)) {
        printf("aiot_gateway_batch_delete_topo_async success %d\r\n", res);
    } else {
        printf("aiot_gateway_batch_delete_topo_async error %d\r\n", res);
        demo_gateway_device_deinit(gateway_device);
        demo_sub_device_deinit(subdev_table, subdev_num);
        return -1;
    }

    /* 删除子设备对象 */
    demo_sub_device_deinit(subdev_table, subdev_num);

    /* 网关设备断连与删除 */
    demo_gateway_device_deinit(gateway_device);
    return 0;
}
