#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "aiot_device_api.h"
#include "aiot_gateway_api.h"
#include "aiot_message_api.h"
#include "demo_config.h"

/* 子设备的认证信息 */
aiot_subdev_meta_info_t sub_device_meta[] = SUBDEV_DYNREG_LIST;

/* 设备的状态回调示例 */
void demo_status_callback(void *device, const aiot_device_status_t *status, void *userdata)
{
    char *product_key = aiot_device_product_key(device);
    char *device_name = aiot_device_name(device);
    if(device == NULL) {
        return;
    }

    switch (status->type) {
    /* SDK因为用户调用了aiot_device_connect()接口, 与mqtt服务器建立连接已成功 */
    case AIOT_DEVICE_STATUS_CONNECT: {
        printf("[%s][%s] AIOT_DEVICE_STATUS_CONNECT\n", product_key, device_name);
        /* TODO: 处理SDK建连成功, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;

    /* SDK因为网络状况被动断连后, 自动发起重连已成功 */
    case AIOT_DEVICE_STATUS_RECONNECT: {
        printf("[%s][%s] AIOT_DEVICE_STATUS_RECONNECT\n", product_key, device_name);
        /* TODO: 处理SDK重连成功, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;

    /* SDK因为网络的状况而被动断开了连接, network是底层读写失败, heartbeat是没有按预期得到服务端心跳应答 */
    case AIOT_DEVICE_STATUS_DISCONNECT: {
        printf("[%s][%s] AIOT_DEVICE_STATUS_DISCONNECT: %d\n", product_key, device_name, status->error_code);
        /* TODO: 处理SDK被动断连, 不可以在这里调用耗时较长的阻塞函数 */
    }
    break;
    }
}

/* 设备的消息回调示例，默认只做打印处理 */
void demo_msg_callback(void *device, const aiot_msg_t *message, void *userdata)
{
    char *product_key = aiot_device_product_key(device);
    char *device_name = aiot_device_name(device);
    if(device == NULL) {
        return;
    }

    printf("[%s][%s] <<, topic: %s\n", product_key, device_name, message->topic);
    printf("[%s][%s] <<, payload: %.*s\n", product_key, device_name, message->payload_lenth, message->payload);
}

/* 主动发送的消息，或者订阅/取消订阅消息有结果后会回调该消息 */
void demo_result_callback(void *device, const aiot_device_msg_result_t *result, void *userdata)
{
    char *product_key = aiot_device_product_key(device);
    char *device_name = aiot_device_name(device);
    if(device == NULL) {
        return;
    }

    switch(result->type) {
        case AIOT_DEVICE_SEND_MESSAGE_RESULT: {
            printf("[%s][%s] send message result, message id: %d, code -0x%04X\n", product_key, device_name, result->message_id, -result->code);
        }
        break;  
        case AIOT_DEVICE_REGISTER_TOPIC_RESULT:{
            printf("[%s][%s] register, message id: %d, code code -0x%04X\n", product_key, device_name, result->message_id, -result->code);
        }
        break;  
        default:
            break;
    }
}

static void* demo_gateway_device_init(char *product_key, char *device_name, char *device_secret, char *host, uint16_t port)
{
    int32_t res = STATE_SUCCESS;
    /* 创建设备 */
    void *device = aiot_device_create(product_key, device_name);
    if (device == NULL) {
        printf("device create failed\n");
        return NULL;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(device, device_secret);

    /* 连接配置参数初始化 */
    aiot_linkconfig_t* config = aiot_linkconfig_init(protocol);
    /* 设置服务器的host、port */
    aiot_linkconfig_host(config, host, port);

    /* 设置设备连接参数 */
    aiot_device_set_linkconfig(device, config);
    
    /* 设备建连 */
    res = aiot_device_connect(device);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_linkconfig_deinit(&config);
        aiot_device_delete(&device);
        printf("aiot_device_connect failed: -0x%04X\n\r\n", -res);
        return NULL;
    }

    /* 建连成功返回设备对象 */
    aiot_linkconfig_deinit(&config);
    return device;
}

static void demo_gateway_device_deinit(void *device)
{
    int32_t res = STATE_SUCCESS;
    res = aiot_device_disconnect(device);
    if (res < STATE_SUCCESS) {
        printf("aiot_device_disconnect failed: -0x%04X\n", -res);
    }
    aiot_device_delete(&device);
}


int main(int argc, char *argv[])
{
    int32_t     res = STATE_SUCCESS;
    int32_t subdev_num = 0;
    /* 配置SDK的底层依赖 */
    aiot_sysdep_init(aiot_sysdep_get_portfile());

    /* 网关设备完成初始化，启动建连，如果失败返回NULL */
    void *gateway_device = demo_gateway_device_init(gateway_product_key, gateway_device_name, gateway_device_secret, host, port);
    if(gateway_device == NULL) {
        return -1;
    }

    subdev_num = sizeof(sub_device_meta) / sizeof(aiot_subdev_meta_info_t);
    /* 子设备动态获取密钥，返回密钥直接写进sub_device_meta中 */
    res = aiot_gateway_dynamic_secret(gateway_device, sub_device_meta, subdev_num);
    if(res < STATE_SUCCESS) {
        printf("aiot_gateway_dynamic_secret error %d\r\n", res);
        return -1;
    }
    /* 子设备动态注册，返回密钥直接写进sub_device_meta中 */
    res = aiot_gateway_dynamic_register(gateway_device, sub_device_meta, subdev_num);
    if(res < STATE_SUCCESS) {
        printf("aiot_gateway_dynamic_register error %d\r\n", res);
        return -1;
    }

    /* 网关设备断连与删除资源 */
    demo_gateway_device_deinit(gateway_device);
    return 0;
}
