#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "aiot_device_api.h"
#include "aiot_logpost_api.h"
#include "demo_config.h"

/* 设备收到OTA的消息 */
static void demo_logpost_msg_callback(void *device, const aiot_logpost_event_t *event, void *userdata)
{
    if(event->type == AIOT_LOGPOSTEVT_CONFIG_DATA) {
        if(event->data.config_data.on_off == 1) {
            printf("device logpost config on\r\n");
        } else {
            printf("device logpost config off\r\n");
        }
    } else if(event->type == AIOT_LOGPOSTEVT_POST_REPLY) {
        printf("device logpost post reply id %d, code %d\r\n", 
            event->data.post_reply.msg_id, event->data.post_reply.code);
    }
}

/* 上报日志到云端 */
static void demo_send_log(void *device, char *log)
{
    int32_t res = 0;
    aiot_log_t log_msg;

    memset(&log_msg, 0, sizeof(aiot_log_t));
    log_msg.timestamp = 0;                          /* 单位为ms的时间戳, 填写0则SDK将使用当前的时间戳 */
    log_msg.loglevel = AIOT_LOGPOST_LEVEL_DEBUG;    /* 日志级别 */
    log_msg.module_name = "APP";                    /* 日志对应的模块 */
    log_msg.code = 200;                             /* 状态码 */
    log_msg.msg_id = 0;                             /* 云端下行报文的消息标示符, 若无对应消息可直接填0 */
    log_msg.content = log;                          /* 日志内容 */

    res = aiot_device_logpost_send(device, &log_msg);
    if (res < 0) {
        printf("aiot_device_logpost_send failed: -0x%04X\r\n", -res);
    }
}

static void* demo_device_init(char *product_key, char *device_name, char *device_secret, char *host, uint16_t port)
{
    int32_t res = STATE_SUCCESS;
    /* 创建设备 */
    void *device = aiot_device_create(product_key, device_name);
    if (device == NULL) {
        printf("device create failed\n");
        return NULL;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(device, device_secret);

    /* 连接配置参数初始化 */
    aiot_linkconfig_t* config = aiot_linkconfig_init(protocol);
    /* 设置服务器的host、port */
    aiot_linkconfig_host(config, host, port);

    /* 设置设备连接参数 */
    aiot_device_set_linkconfig(device, config);
    
    /* 设备建连 */
    res = aiot_device_connect(device);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_linkconfig_deinit(&config);
        aiot_device_delete(&device);
        printf("aiot_device_connect failed: -0x%04X\n\r\n", -res);
        return NULL;
    }

    /* 建连成功返回设备对象 */
    aiot_linkconfig_deinit(&config);
    return device;
}

static void demo_device_deinit(void *device)
{
    int32_t res = STATE_SUCCESS;
    res = aiot_device_disconnect(device);
    if (res < STATE_SUCCESS) {
        printf("aiot_device_disconnect failed: -0x%04X\n", -res);
    }
    aiot_device_delete(&device);
}

int main(int argc, char *argv[])
{
    /* 配置SDK的底层依赖 */
    aiot_sysdep_init(aiot_sysdep_get_portfile());

    /* 初始化设备，并将设备连接上云 */
    void *device = demo_device_init(product_key, device_name, device_secret, host, port);
    if(device == NULL) {
        printf("demo device connect failed\r\n");
    }

    /* 设置日志模块回调函数 */
    aiot_device_logpost_set_callback(device, demo_logpost_msg_callback, NULL);
    /* 设置日志上报是否需要回复 0:默认值，不需要回复 1.需要回复 */
    aiot_device_logpost_set_post_reply(device, 1);

    int i = 5;
    while(i--) {
        demo_send_log(device, "logpost test in whilt(1)");
        sleep(10);
    }

    /* 断开设备连接，并回收设备资源 */
    demo_device_deinit(device);
    return 0;
}
