#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include "utils_md5.h"
#include "http_download.h"
#include "aiot_device_api.h"
#include "aiot_ota_api.h"
#include "demo_config.h"

typedef struct {
    /* 设备句柄 */
    void *device;
    /* ota下载文件标志，避免重复下载，非0表示升级中 */
    int32_t ota_process_flag;
    /* 拷贝的OTA消息 */
    aiot_ota_msg_t *ota_msg;
    /* 保存的文件名 */
    char *filename;
} demo_ota_task_t;

/* 接收到的固件数据保存文件, 该接入如果返回负数就会中断下载流程 */
static int32_t save_file(const char *filename, uint32_t offset, uint8_t *data, uint32_t data_len, void *userdata)
{
    FILE *file = NULL;
    printf("recv data offset %d, len %d\r\n", offset, data_len);
    if(filename == NULL) {
        return -1;
    }

    /* 保存一个新文件时，有同名文件时，会先删除旧的文件数据 */
    if(offset == 0 ) {
        file = fopen(filename, "wb");
    } else {
        file = fopen(filename, "ab");
    }

    fwrite(data, sizeof(int8_t), data_len, file);
    fclose(file);

    return data_len;
}

/* hex转字符串，用于md5码计算 */
int32_t hex2str(uint8_t *input, uint32_t input_len, char *output, uint8_t lowercase)
{
    char *upper = "0123456789ABCDEF";
    char *lower = "0123456789abcdef";
    char *encode = upper;
    int i = 0, j = 0;

    if (lowercase) {
        encode = lower;
    }

    for (i = 0; i < input_len; i++) {
        output[j++] = encode[(input[i] >> 4) & 0xf];
        output[j++] = encode[(input[i]) & 0xf];
    }

    return STATE_SUCCESS;
}

/* 计算文件的md5码 */
static int32_t file_md5_check(char *filename, char *md5_digest)
{
    FILE *file = NULL;
    uint8_t data[2048];
    int32_t read_len = 0;
    utils_md5_context_t ctx;
    uint8_t md5_calc[16] = {0};
    char md5_calc_string[33] = {0};
    if(filename == NULL || md5_digest == NULL) {
        return -1;
    }

    utils_md5_init(&ctx);
    utils_md5_starts(&ctx);
    file = fopen(filename, "rb+");
    while((read_len = fread(data, sizeof(int8_t), 2048, file)) > 0){
        printf("update size %d\r\n",read_len);
        utils_md5_update(&ctx, data, read_len);
    }
    fclose(file);
    utils_md5_finish(&ctx, md5_calc);
    hex2str(md5_calc, sizeof(md5_calc), md5_calc_string, 1);

    printf("calc %s digest %s\r\n", md5_calc_string, md5_digest);

    if (memcmp(md5_calc_string, md5_digest, sizeof(md5_calc_string)) == 0) {
        return STATE_SUCCESS;
    } else {
        return -1;
    }
}

/* 处理升级流程 */
static void *demo_ota_process_thread(void *args)
{
    demo_ota_task_t *task = (demo_ota_task_t *)args;
    int32_t res = STATE_SUCCESS;
    http_download_params_t params = HTTP_DOWNLOAD_PARAMS_DEFAULT();
    printf("version %s, http_url %s\r\n", task->ota_msg->version, task->ota_msg->http_url);

    task->ota_process_flag = 1;
    /* 下载文件 */
    res = core_http_download_request(task->ota_msg->http_url, params, task->filename, save_file, NULL);
    if(res != STATE_SUCCESS) {
        printf("core_http_download_request failed\r\n");
        /* 下载失败 */
        aiot_device_ota_report_state(task->device, NULL, -2);
        task->ota_process_flag = 0;
        return NULL;
    } 

    printf("core_http_download_request success, digest %s \r\n", task->ota_msg->expect_digest);
    
    /* http下载如果是整个文件下载，已经校验过下载内容，此处可以添加文件系统内文件校验，更加保险 */
    if(task->ota_msg->digest_method == AIOT_OTA_DIGEST_MD5) {
        res = file_md5_check(task->filename, task->ota_msg->expect_digest);
        if(STATE_SUCCESS != res ) {
            /* 校验固件失败，上报状态 */
            aiot_device_ota_report_state(task->device, NULL, -3);
        }
    }

    /* 固件校验成功 */
    if(res == STATE_SUCCESS) {
        /* 烧写固件，如果烧写失败，上报错误状态 */
        /*
            aiot_device_ota_report_state(task->device, NULL, -4);
        */

        /* 写文件系统或者flash成功，上报升级进度 */
        aiot_device_ota_report_state(task->device, NULL, 100);
        
        /* 更新新版本号表示升级完成，建议是重启后再上报 */
        /*
        aiot_device_ota_report_version(task->device, NULL, ota_msg->version);
        */
    }

   /* 删除ota_msg */
   aiot_ota_msg_free(task->ota_msg);
   task->ota_msg = NULL;
   task->ota_process_flag = 0;

   return NULL;
}

/* 设备接收收到OTA消息的回调函数 */
static void demo_ota_msg_callback(void *device, const aiot_ota_msg_t *ota_msg, void *userdata)
{
    int32_t res = STATE_SUCCESS;
    pthread_t ota_process_thread;
    demo_ota_task_t *task = (demo_ota_task_t *)userdata;
    if(ota_msg != NULL && ota_msg->http_url != NULL) {
        /* 判断设备是否已经在执行升级，避免重复升级 */
        if(task->ota_process_flag != 0) {
            return;
        }

        /* 回调中不能长时间阻塞，拷贝消息用于异步处理, 拷贝的对象需要删除 */
        task->ota_msg = aiot_ota_msg_clone(ota_msg);
        res = pthread_create(&ota_process_thread, NULL, demo_ota_process_thread, task);
        if (res < 0) {
            aiot_ota_msg_free(task->ota_msg);
            task->ota_msg = NULL;
            printf("pthread_create demo_ota_process_thread failed: %d\n", res);
            return;
        } else {
            /* 设置线程为detach状态，用于自动资源回收 */
            pthread_detach(ota_process_thread);
        }
    } else {
        printf("ota invalid message \r\n");
    }
}

/* 初始化设备句柄，并且完成设备建连 */
static void* demo_device_init(char *product_key, char *device_name, char *device_secret, char *host, uint16_t port)
{
    int32_t res = STATE_SUCCESS;
    /* 创建设备 */
    void *device = aiot_device_create(product_key, device_name);
    if (device == NULL) {
        printf("device create failed\n");
        return NULL;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(device, device_secret);

    /* 连接配置参数初始化 */
    aiot_linkconfig_t* config = aiot_linkconfig_init(protocol);
    /* 设置服务器的host、port */
    aiot_linkconfig_host(config, host, port);

    /* 设置设备连接参数 */
    aiot_device_set_linkconfig(device, config);
    
    /* 设备建连 */
    res = aiot_device_connect(device);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_linkconfig_deinit(&config);
        aiot_device_delete(&device);
        printf("aiot_device_connect failed: -0x%04X\n\r\n", -res);
        return NULL;
    }

    /* 建连成功返回设备对象 */
    aiot_linkconfig_deinit(&config);
    return device;
}

/* 断开连接，删除设备，回收设备资源 */
static void demo_device_deinit(void *device)
{
    int32_t res = STATE_SUCCESS;
    res = aiot_device_disconnect(device);
    if (res < STATE_SUCCESS) {
        printf("aiot_device_disconnect failed: -0x%04X\n", -res);
    }
    aiot_device_delete(&device);
}

int main(int argc, char *argv[])
{
    /* 示例定义升级任务，用于ota过程管理 */
    demo_ota_task_t ota_task;
    
    /* 配置SDK的底层依赖 */
    aiot_sysdep_init(aiot_sysdep_get_portfile());

    /* 初始化设备，并将设备连接上云 */
    void *device = demo_device_init(product_key, device_name, device_secret, host, port);
    if(device == NULL) {
        printf("demo device connect failed\r\n");
    }

    memset(&ota_task, 0, sizeof(ota_task));
    ota_task.device = device;
    ota_task.ota_process_flag = 0;
    ota_task.filename = "fireware.bin";
    /* 设置OTA回调函数 */
    aiot_device_ota_set_callback(device, demo_ota_msg_callback, &ota_task);

    /* 上报设备版本号 */
    aiot_device_ota_report_version(device, NULL, "4.0.0");
    /* 设备主动请求升级包，如果有升级任务，会下发 */
    aiot_device_ota_request_firmware(device, NULL);

    while(1) {
        sleep(10);
    }

    /* 断开设备连接，并回收设备资源 */
    demo_device_deinit(device);
    return 0;
}
