#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <stdlib.h>
#include "aiot_device_api.h"
#include "aiot_rc_api.h"
#include "demo_config.h"
#include "http_download.h"

typedef struct {
    /* 保存设备句柄 */
    void *device;
    /* 文件签名，可用于文件校验 */
    char *sign;
    /* 文件下载地址 */
    char *url;
    /* 文件保存在本地的地址 */
    char *filename;
} demo_config_task_t;

static void* demo_device_init(char *product_key, char *device_name, char *device_secret, char *host, uint16_t port)
{
    int32_t res = STATE_SUCCESS;
    /* 创建设备 */
    void *device = aiot_device_create(product_key, device_name);
    if (device == NULL) {
        printf("device create failed\n");
        return NULL;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(device, device_secret);

    /* 连接配置参数初始化 */
    aiot_linkconfig_t* config = aiot_linkconfig_init(protocol);
    /* 设置服务器的host、port */
    aiot_linkconfig_host(config, host, port);

    /* 设置设备连接参数 */
    aiot_device_set_linkconfig(device, config);
    
    /* 设备建连 */
    res = aiot_device_connect(device);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_linkconfig_deinit(&config);
        aiot_device_delete(&device);
        printf("aiot_device_connect failed: -0x%04X\n\r\n", -res);
        return NULL;
    }

    /* 建连成功返回设备对象 */
    aiot_linkconfig_deinit(&config);
    return device;
}

static void demo_device_deinit(void *device)
{
    int32_t res = STATE_SUCCESS;
    res = aiot_device_disconnect(device);
    if (res < STATE_SUCCESS) {
        printf("aiot_device_disconnect failed: -0x%04X\n", -res);
    }
    aiot_device_delete(&device);
}

static int32_t save_file(const char *filename, uint32_t offset, uint8_t *data, uint32_t data_len, void *userdata)
{
    FILE *file = NULL;
    printf("recv data offset %d, len %d\r\n", offset, data_len);
    if(filename == NULL) {
        return -1;
    }

    /* 保存一个新文件时，有同名文件时，会先删除旧的文件数据 */
    if(offset == 0 ) {
        file = fopen(filename, "wb");
    } else {
        file = fopen(filename, "ab");
    }

    /* 用户应在此实现文件本地固化的操作 */
    fwrite(data, sizeof(int8_t), data_len, file);
    fclose(file);

    return data_len;
}

/* 下载远程配置文件,并更新设置 */
static void *demo_remote_config_process_thread(void *args)
{
    demo_config_task_t *task = (demo_config_task_t *)args;
    int32_t res = STATE_SUCCESS;
    http_download_params_t params = HTTP_DOWNLOAD_PARAMS_DEFAULT();

    /* 下载文件 */
    res = core_http_download_request(task->url, params, task->filename, save_file, NULL);
    if(res != STATE_SUCCESS) {
        printf("core_http_download_request failed\r\n");
        return NULL;
    }
    printf("remote config file %s download success\r\n", task->filename);

    /* 资源回收退出 */
    free(task->url);
    free(task->sign);
    return NULL;
}

/* 远程配置回调处理 */
static void demo_rc_callback(void *device, const aiot_rc_msg_t *rc_msg, void *userdata)
{
    demo_config_task_t *task = (demo_config_task_t *)userdata;
    int32_t res = STATE_SUCCESS;
    pthread_t config_process_thread;
    if(device == NULL || rc_msg == NULL) {
        return;
    }

    printf("file size %d, url :%s\r\n", rc_msg->config_size, rc_msg->url);

    /* 正在执行配置任务的话，直接退出 */
    if(task->url != NULL) {
        return;
    }

    task->url = strdup(rc_msg->url);
    task->sign = strdup(rc_msg->sign);
    res = pthread_create(&config_process_thread, NULL, demo_remote_config_process_thread, task);
    if (res < 0) {
        free(task->url);
        free(task->sign);
        printf("pthread_create demo_rc_callback failed: %d\n", res);
        return;
    } else {
        /* 设置线程为detach状态，用于自动资源回收 */
        pthread_detach(config_process_thread);
    }
}

int main(int argc, char *argv[])
{
    void *device = NULL;
    demo_config_task_t config_task;
    /* 配置SDK的底层依赖 */
    aiot_sysdep_init(aiot_sysdep_get_portfile());

    /* 初始化设备，并将设备连接上云 */
    device = demo_device_init(product_key, device_name, device_secret, host, port);
    if(device == NULL) {
        printf("demo device connect failed\r\n");
    }

    memset(&config_task, 0, sizeof(config_task));
    config_task.device = device;
    config_task.filename = "demo_config.json";
    config_task.url = NULL;
    config_task.sign = NULL;
    /* 设置远程配置回调函数 */
    aiot_device_rc_set_callback(device, demo_rc_callback, &config_task);

    /* 主动请求远程配置文件 */
    aiot_device_rc_request(device);
    int i = 5;
    while(i--) {
        sleep(10);
    }

    /* 断开设备连接，并回收设备资源 */
    demo_device_deinit(device);
    return 0;
}
