#include <stdio.h>
#include <unistd.h>
#include "aiot_device_api.h"
#include "aiot_shadow_api.h"
#include "demo_config.h"

/* 设备收到shadow的消息 */
static void demo_shadow_recv_callback(void *device, const aiot_shadow_recv_t *recv, void *userdata)
{
    char *product_key = aiot_device_product_key(device);
    char *device_name = aiot_device_name(device);
    printf("demo_shadow_recv_callback, type = %d, productKey = %s, deviceName = %s\r\n",
            recv->type, product_key, device_name);

    switch (recv->type) {
        /* 当设备发送AIOT_SHADOWMSG_UPDATE, AIOT_SHADOWMSG_CLEAN_DESIRED或者AIOT_SHADOWMSG_DELETE_REPORTED消息后, 会收到此应答消息 */
        case AIOT_SHADOWRECV_GENERIC_REPLY: {
            const aiot_shadow_recv_generic_reply_t *generic_reply = &recv->data.generic_reply;

            printf("payload = \"%.*s\", status = %s, timestamp = %ld\r\n",
                   generic_reply->payload_len,
                   generic_reply->payload,
                   generic_reply->status,
                   (unsigned long)generic_reply->timestamp);
        }
        break;
        /* 当设备在线时, 若用户应用程序调用云端API主动更新设备影子, 设备便会收到此消息 */
        case AIOT_SHADOWRECV_CONTROL: {
            const aiot_shadow_recv_control_t *control = &recv->data.control;

            printf("payload = \"%.*s\", version = %ld\r\n",
                   control->payload_len,
                   control->payload,
                   (unsigned long)control->version);
        }
        break;
        /* 当设备发送AIOT_SHADOWMSG_GET消息主动获取设备影子时, 会收到此消息 */
        case AIOT_SHADOWRECV_GET_REPLY: {
            const aiot_shadow_recv_get_reply_t *get_reply = &recv->data.get_reply;

            printf("payload = \"%.*s\", version = %ld\r\n",
                   get_reply->payload_len,
                   get_reply->payload,
                   (unsigned long)get_reply->version);
        }
        default:
            break;
    }
}

static void* demo_device_init(char *product_key, char *device_name, char *device_secret, char *host, uint16_t port)
{
    int32_t res = STATE_SUCCESS;
    /* 创建设备 */
    void *device = aiot_device_create(product_key, device_name);
    if (device == NULL) {
        printf("device create failed\n");
        return NULL;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(device, device_secret);

    /* 连接配置参数初始化 */
    aiot_linkconfig_t* config = aiot_linkconfig_init(protocol);
    /* 设置服务器的host、port */
    aiot_linkconfig_host(config, host, port);

    /* 设置设备连接参数 */
    aiot_device_set_linkconfig(device, config);
    
    /* 设备建连 */
    res = aiot_device_connect(device);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_linkconfig_deinit(&config);
        aiot_device_delete(&device);
        printf("aiot_device_connect failed: -0x%04X\n\r\n", -res);
        return NULL;
    }

    /* 建连成功返回设备对象 */
    aiot_linkconfig_deinit(&config);
    return device;
}

static void demo_device_deinit(void *device)
{
    int32_t res = STATE_SUCCESS;
    res = aiot_device_disconnect(device);
    if (res < STATE_SUCCESS) {
        printf("aiot_device_disconnect failed: -0x%04X\n", -res);
    }
    aiot_device_delete(&device);
}

int main(int argc, char *argv[])
{
    /* 配置SDK的底层依赖 */
    aiot_sysdep_init(aiot_sysdep_get_portfile());

    /* 初始化设备，并将设备连接上云 */
    void *device = demo_device_init(product_key, device_name, device_secret, host, port);
    if(device == NULL) {
        printf("demo device connect failed\r\n");
    }

    /* 设置日志模块回调函数 */
    aiot_device_shadow_set_callback(device, demo_shadow_recv_callback, NULL);

    int i = 1;
    while(i--) {
        /* 发送请求拉取设备影子, 设备影子将在demo_shadow_recv_callback回调函数中返回 */
        aiot_device_shadow_get(device);
        sleep(2);

        /* 清除设备影子中的desired值, 同时将version设置为1 */
        aiot_device_shadow_clean_desired(device, 1);
        sleep(2);

        /* 删除设备影子中特定的reported值, 同时将version设置为2 */
        aiot_device_shadow_delete_reported(device, "{\"LightSwitch\":\"null\"}", 2);
        sleep(2);

        /* 更新设备影子中的reported值, 同时将version重置为0 */
        aiot_device_shadow_update(device, "{\"LightSwitch\":1}", 0);
        
        sleep(5);
    }

    /* 断开设备连接，并回收设备资源 */
    demo_device_deinit(device);
    return 0;
}
