#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "aiot_device_api.h"
#include "aiot_tunnel_switch_api.h"
#include "aiot_tunnel_api.h"
#include "demo_config.h"

/* 隧道代理服务参数，示例为ssh服务(_SSH)参数，可以用于远程登录 */
aiot_tunnel_proxy_params_t ssh_service = {
    .ip = "127.0.0.1",
    .port = 22,
};

/* 隧道的句柄 */
void *tunnel_handle = NULL;
/* 隧道的状态 */
int8_t tunnel_status = -1;

static void* demo_device_init(char *product_key, char *device_name, char *device_secret, char *host, uint16_t port)
{
    int32_t res = STATE_SUCCESS;
    /* 创建设备 */
    void *device = aiot_device_create(product_key, device_name);
    if (device == NULL) {
        printf("device create failed\n");
        return NULL;
    }
    /* 设置设备密钥 */
    aiot_device_set_device_secret(device, device_secret);

    /* 连接配置参数初始化 */
    aiot_linkconfig_t* config = aiot_linkconfig_init(protocol);
    /* 设置服务器的host、port */
    aiot_linkconfig_host(config, host, port);

    /* 设置设备连接参数 */
    aiot_device_set_linkconfig(device, config);

    /* 设备建连 */
    res = aiot_device_connect(device);
    if (res < STATE_SUCCESS) {
        /* 尝试建立连接失败, 销毁MQTT实例, 回收资源 */
        aiot_linkconfig_deinit(&config);
        aiot_device_delete(&device);
        printf("aiot_device_connect failed: -0x%04X\n\r\n", -res);
        return NULL;
    }

    /* 建连成功返回设备对象 */
    aiot_linkconfig_deinit(&config);
    return device;
}

static void demo_device_deinit(void *device)
{
    int32_t res = STATE_SUCCESS;
    res = aiot_device_disconnect(device);
    if (res < STATE_SUCCESS) {
        printf("aiot_device_disconnect failed: -0x%04X\n", -res);
    }
    aiot_device_delete(&device);
}

/* 隧道的事件回调 */
void demo_tunnel_event_cb(void *handle, const aiot_tunnel_event_type event, void *userdata)
{
    switch(event)
    {
    case AIOT_TUNNEL_EVT_CONNECT:
        printf( "demo_tunnel_event_cb AIOT_TUNNEL_EVT_CONNECT \r\n");
        /* TODO: 告知websocket建连成功, 不可以在这里调用耗时较长的阻塞函数 */
        break;
    case AIOT_TUNNEL_EVT_DISCONNECT:
        printf( "demo_tunnel_event_cb AIOT_TUNNEL_EVT_DISCONNECT \r\n");
        /* TODO: 告知websocket掉线, 不可以在这里调用耗时较长的阻塞函数 */
        break;
    case AIOT_TUNNEL_EVT_EXPIRED:
        printf( "demo_tunnel_event_cb AIOT_TUNNEL_EVT_EXPIRED \r\n");
        /* TODO: 隧道认证信息过期，需要更新隧道验证信息或者重新打开隧道 */
        break;
    }

    tunnel_status = event;
}

/* 设备的安全隧道开关模块回调，处理云端信令 */
static void demo_tunnel_switch_callback(void *device, const aiot_ts_msg_t *ts_msg, void *userdata)
{
    aiot_tunnel_connect_param_t params;
    if(device == NULL || ts_msg == NULL) {
        return;
    }
    printf("tunnel msg recv: tunnel_id %s, will expired after %ds\r\n", ts_msg->tunnel_id, ts_msg->expired_time);

    /* 该示例只有一个隧道，如果为关闭隧道命令则关闭隧道，如果为打开新隧道，也需要关闭旧的隧道*/
    if(tunnel_handle != NULL) {
        aiot_tunnel_deinit(&tunnel_handle);
        tunnel_status = -1;
    }

    if(AIOT_TUNNEL_OPERATOPN_OPEN == ts_msg->operation ||
            AIOT_TUNNEL_OPERATOPN_UPDATE == ts_msg->operation) {
        memset(&params, 0, sizeof(params));
        params.host = ts_msg->host;
        params.port = ts_msg->port;
        params.path = ts_msg->path;
        params.token = ts_msg->token;

        /* 创建并初始化新的隧道 */
        tunnel_handle = aiot_tunnel_init(ts_msg->tunnel_id);
        /* 配置隧道支持的服务 */
        aiot_tunnel_add_proxy_service(tunnel_handle, "_SSH", &ssh_service);
        /* 配置隧道状态的回调函数 */
        aiot_tunnel_set_event_callback(tunnel_handle, demo_tunnel_event_cb, NULL);
        /* 设置隧道的连接参数 */
        aiot_tunnel_set_connect_params(tunnel_handle, &params);

        /* 回调函数中不能长时间阻塞，在主函数中完成建连 */
    }
}

int main(int argc, char *argv[])
{
    void *device = NULL;
    /* 配置SDK的底层依赖 */
    aiot_sysdep_init(aiot_sysdep_get_portfile());

    /* 初始化设备，并将设备连接上云 */
    device = demo_device_init(product_key, device_name, device_secret, host, port);
    if(device == NULL) {
        printf("demo device connect failed\r\n");
    }

    /* 设置隧道开关消息回调 */
    aiot_device_ts_set_callback(device, demo_tunnel_switch_callback, NULL);

    /* 主动请求隧道信息 */
    aiot_device_ts_request(device);

    while(1) {
        if(tunnel_handle != NULL) {
            if(tunnel_status == -1) {
                /* 隧道未建连，异步完成隧道建连,在事件回调函数中重置tunnel_status状态 */
                aiot_tunnel_connect(tunnel_handle);
            } else if(tunnel_status == AIOT_TUNNEL_EVT_EXPIRED) {
                /* 隧道已经过期，删除过期隧道 */
                aiot_tunnel_disconnect(tunnel_handle);
                aiot_tunnel_deinit(&tunnel_handle);
                tunnel_status = -1;
            }
        }

        sleep(1);
    }

    /* 断开设备连接，并回收设备资源 */
    demo_device_deinit(device);

    if(tunnel_handle != NULL) {
        aiot_tunnel_disconnect(tunnel_handle);
        aiot_tunnel_deinit(&tunnel_handle);
        tunnel_status = -1;
    }

    return 0;
}
